/*
 *
 * Copyright (C) 2005 Yoann Vandoorselaere, Prelude IDS Technologies
 *                    Rainer Wichmann
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/*
 * 28/04/2005 : R.W.:
 *       move libprelude 0.8 code to seperate file
 *
 * 23/04/2005 : R.W.: 
 *       include libprelude 0.9 code from Yoann Vandoorselaere
 */
 
#include "config_xor.h"

#include <stdio.h>
#include <string.h>
#include <sys/types.h>

#if TIME_WITH_SYS_TIME
#include <sys/time.h>
#include <time.h>
#else
#if HAVE_SYS_TIME_H
#include <sys/time.h>
#else
#include <time.h>
#endif
#endif
#include <unistd.h>
#include <syslog.h>
#include <pwd.h>
#include <grp.h>

#if defined(HAVE_LIBPRELUDE) && defined(HAVE_LIBPRELUDE_9)


/*
 * _() macros are samhain specific; they are used to replace string
 * constants at runtime. This is part of the samhain stealth mode
 * (fill string constants with encoded strings, decode at runtime).
 */
#define FIL__  _("sh_prelude.c")


#include <libprelude/idmef.h>
#include <libprelude/prelude.h>

/* 
 * includes for samhain-specific functions (sl_strstr, sh_error_handle)
 */
#include "slib.h"
#include "sh_cat.h"
#include "sh_error_min.h"
#include "sh_prelude.h"

#define IDMEF_ANALYZER_MODEL _("Samhain")
#define IDMEF_ANALYZER_CLASS _("Integrity Checker")
#define IDMEF_ANALYZER_VERSION VERSION
#define IDMEF_ANALYZER_MANUFACTURER _("Samhain by Rainer Wichmann")

#define CLASSIFICATION_URL _("http://www.la-samhna.de/samhain/")

typedef enum {
        S_MODIFICATION_CHECKSUM  = (1 << 0),
        S_MODIFICATION_LINK      = (1 << 1),
        S_MODIFICATION_INODE     = (1 << 2),
        S_MODIFICATION_USER      = (1 << 3),
        S_MODIFICATION_GROUP     = (1 << 4),
        S_MODIFICATION_TIME      = ((1 << 5)|(1 << 6)|(1 << 7)),
        S_MODIFICATION_SIZE      = (1 << 8),
        S_MODIFICATION_MODE      = (1 << 9),
        S_MODIFICATION_HARDLINKS = (1 << 10),
        S_MODIFICATION_DEVICE    = (1 << 11),
} s_modification_t;


static prelude_client_t *client = NULL;
static char sh_prelude_model[32];
static char sh_prelude_class[32];
static char sh_prelude_version[32];
/* static char sh_prelude_manufacturer[32]; */

/* returns 0/tiger, 1/sha1, or 2/md5
 */
extern int sh_tiger_get_hashtype();


/* Get the value for a key. The key is built from toktmp + toksuffix.
 * msg is modified temporarily, so it should not be declared 'const'.
 */
static char *get_value (char *msg, const char *toktmp, 
			const char *toksuffix)
{
  char * ret = NULL, *ptr, tok[128];

  snprintf(tok, sizeof(tok), "%s%s", toktmp, (toksuffix) ? toksuffix : "");

  ptr = strstr(msg, tok);
  if ( ! ptr )
    return NULL;

#ifdef SH_USE_XML
  while (*ptr && *ptr != '"') ptr++;
  if (*ptr) { 
    ret = ptr + 1; ptr = ret; 
  }
  while (*ptr && *ptr != '"') ptr++;
  if (*ptr)
    {
      *ptr = '\0';
      if (ret) ret = strdup(ret);
      *ptr = '"';
    }
  else
    {
      if (ret) ret = strdup(ret);
    }
#else
  while (*ptr && *ptr != '<') ptr++;
  if (*ptr) ret = ptr + 1;
  while (*ptr && *ptr != '>') ptr++;
  if (*ptr)
    {
      *ptr = '\0';
      if (ret) ret = strdup(ret);
      *ptr = '>';
    }
  else
    {
      if (ret) ret = strdup(ret);
    }
#endif

  return ret;
}

static void get_file_infos(idmef_target_t *target, char *msg, 
			   idmef_file_category_t category)
{
        int ret;
        int hashtype = 0;
        char *ptr;
        idmef_time_t *time;
        idmef_file_t *file;
        idmef_inode_t *inode;
        prelude_string_t *str;
        idmef_checksum_t *checksum;
        idmef_file_access_t *access;
        idmef_user_id_t *userid;
        const char *suffix = (category == IDMEF_FILE_CATEGORY_CURRENT) ? "_new" : "_old";
                
        ret = idmef_target_new_file(target, &file, -1);
        if ( ret < 0  )
                return;
        idmef_file_set_category(file, category);

        ptr = get_value(msg, _("path"), NULL);
        if ( ptr ) {
                /*
                 * In term of IDMEF, this is the full path,
                 * including the name.
                 */
                ret = idmef_file_new_path(file, &str);
                if ( ret < 0 )
                        return;
                prelude_string_set_nodup(str, ptr);

                ptr = strrchr(ptr, '/');
                if ( ptr ) {
                        ret = idmef_file_new_name(file, &str);
                        if ( ret < 0 )
                                return;
                        
                        prelude_string_set_dup(str, ptr + 1);
                }
        }
                
        ptr = get_value(msg, _("size"), suffix);
	if ( ptr ) {
                idmef_file_set_data_size(file, strtoul(ptr, NULL, 10));
                free(ptr);
        }
        
        ptr = get_value(msg, _("mtime"), suffix);
        if ( ptr ) {
		ptr[strlen(ptr) - 1] = 0;
                ret = idmef_time_new_from_string(&time, ptr + 1);
                if ( ret < 0 )
                        return;
                
                idmef_file_set_modify_time(file, time);
                free(ptr);
        }

        ptr = get_value(msg, _("ctime"), suffix);
        if ( ptr ) {
                ptr[strlen(ptr) - 1] = 0;
                
                ret = idmef_time_new_from_string(&time, ptr + 1);
                if ( ret < 0 )
                        return;
                
                idmef_file_set_create_time(file, time);
                free(ptr);
        }
                
        ptr = get_value(msg, _("inode"), suffix);
        if ( ptr ) {
                ret = idmef_file_new_inode(file, &inode);
                if ( ret < 0 )
                        return;
                
                idmef_inode_set_number(inode, strtoul(ptr, NULL, 10));
                free(ptr);
        }

        ptr = get_value(msg, _("chksum"), suffix);
        if ( ptr ) {
                ret = idmef_file_new_checksum(file, &checksum, 0);
                if ( ret < 0 )
                        return;

		hashtype = sh_tiger_get_hashtype();

		if (hashtype == 0)
		  idmef_checksum_set_algorithm(checksum, IDMEF_CHECKSUM_ALGORITHM_TIGER);
		else if (hashtype == 1)
		  idmef_checksum_set_algorithm(checksum, IDMEF_CHECKSUM_ALGORITHM_SHA1);
		else if (hashtype == 2)
		  idmef_checksum_set_algorithm(checksum, IDMEF_CHECKSUM_ALGORITHM_MD5);
		else
		  idmef_checksum_set_algorithm(checksum, IDMEF_CHECKSUM_ALGORITHM_TIGER);


                ret = idmef_checksum_new_value(checksum, &str);
                if ( ret < 0 )
                        return;

		/* will be freed on destroy()
		 */
                prelude_string_set_nodup(str, ptr);
        }

        ptr = get_value(msg, _("owner"), suffix);
        if ( ptr ) {
                struct passwd *pw;
                
                ret = idmef_file_new_file_access(file, &access, 0);
                if ( ret < 0 )
                        return;

                ret = idmef_file_access_new_user_id(access, &userid);
                if ( ret < 0 )
                        return;
                idmef_user_id_set_type(userid, IDMEF_USER_ID_TYPE_USER_PRIVS);
                
                ret = idmef_user_id_new_name(userid, &str);
                if ( ret < 0 )
                        return;
                
                prelude_string_set_nodup(str, ptr);
                
                pw = getpwnam(ptr);
                if ( ! pw )
                        return;
                
                idmef_user_id_set_number(userid, pw->pw_uid);
        }

        ptr = get_value(msg, _("group"), suffix);
        if ( ptr ) {
                struct group *gr;
                
                ret = idmef_file_new_file_access(file, &access, -1);
                if ( ret < 0 )
                        return;

                ret = idmef_file_access_new_user_id(access, &userid);
                if ( ret < 0 )
                        return;
                idmef_user_id_set_type(userid, IDMEF_USER_ID_TYPE_GROUP_PRIVS);
                
                ret = idmef_user_id_new_name(userid, &str);
                if ( ret < 0 )
                        return;
                
                prelude_string_set_nodup(str, ptr);

                gr = getgrnam(ptr);
                if ( ! gr )
                        return;

                idmef_user_id_set_number(userid, gr->gr_gid);
        }
}

static int map_policy_to_class(char * msg, unsigned long msgid, prelude_string_t *out)
{
  char * ptr;
  int    ret;

#if defined(SH_WITH_CLIENT) || defined(SH_STANDALONE)
  if (msgid == MSG_FI_MISS || msgid == MSG_FI_MISS2)
    return prelude_string_cat(out, _("File Missing"));

  if (msgid == MSG_FI_ADD || msgid == MSG_FI_ADD2)
    return prelude_string_cat(out, _("File Added"));

  if (msgid == MSG_FI_CHAN)
    return prelude_string_cat(out, _("File Modified"));

   if (msgid == MSG_FI_NODIR)
    return prelude_string_cat(out, _("File found where directory was expected"));
#endif

#ifdef SH_USE_KERN
 if (msgid == MSG_KERN_POLICY || msgid == MSG_KERN_POL_CO || 
      msgid == MSG_KERN_PROC   || msgid == MSG_KERN_GATE ||
      msgid == MSG_KERN_IDT    || msgid == MSG_KERN_SYSCALL)
    return prelude_string_cat(out, _("Kernel Modified"));
#endif
        
#ifdef SH_USE_SUIDCHK
   if (msgid == MSG_SUID_POLICY)
    return prelude_string_cat(out, _("SUID/SGID File Detected"));
#endif

   if (msgid == MSG_EXIT_NORMAL)
    return prelude_string_cat(out, _("Exiting"));

   if (msgid == MSG_START_KEY)
    return prelude_string_cat(out, _("Sending Logfile Key"));

   if (msgid == MSG_START_0H || msgid == MSG_START_1H || 
       msgid == MSG_START_2H || msgid == MSG_START_GH)
    return prelude_string_cat(out, _("Starting Up"));

   if (msgid == MSG_START_GH2)
    return prelude_string_cat(out, _("Startup Failed"));

   if (msgid == MSG_SUSPEND)
    return prelude_string_cat(out, _("Suspending Execution"));

   if (msgid == MSG_RECONF)
    return prelude_string_cat(out, _("Reloading Runtime Configuration"));

   /* some other message
    */
   ptr = get_value(msg, _("msg"), NULL);
   ret = prelude_string_cat(out, ptr);
   free(ptr);
   return ret;
}





static char sh_prelude_client_name[8];

static int sh_prelude_init(void)
{
        int ret;
        prelude_string_t *str;
        idmef_analyzer_t *analyzer;
        prelude_client_flags_t flags;

        prelude_init(NULL, NULL);
        /* prelude_log_set_debug_level(10); */
        
	sl_strlcpy(sh_prelude_client_name, _("Samhain"), 8);
	/* 
	 * Does this copy the second argument or reference it ???
	 */
        ret = prelude_client_new(&client, sh_prelude_client_name);
        if ( ret < 0 ) {
                sh_error_handle((-1), FIL__, __LINE__, -1, MSG_E_SUBGEN,
                                _("Failed to initialize Prelude"), _("sh_prelude_init"));
                return -1;
        }

        /*
         * Enable automatic heartbeat sending.
         */
        flags = prelude_client_get_flags(client);
        ret = prelude_client_set_flags(client, flags | PRELUDE_CLIENT_FLAGS_ASYNC_TIMER);

        ret = prelude_client_start(client);
        if ( ret < 0 ) {
                prelude_perror(ret, _("error starting prelude client"));
                
                sh_error_handle((-1), FIL__, __LINE__, -1, MSG_E_SUBGEN,
                                _("Failed to start Prelude"), _("sh_prelude_init"));
                return -1;
        }
        
        analyzer = prelude_client_get_analyzer(client);

	sl_strlcpy(sh_prelude_model, IDMEF_ANALYZER_MODEL, 32);
        ret = idmef_analyzer_new_model(analyzer, &str);
        prelude_string_set_constant(str, sh_prelude_model);

	sl_strlcpy(sh_prelude_class, IDMEF_ANALYZER_CLASS, 32);
        ret = idmef_analyzer_new_class(analyzer, &str);
        prelude_string_set_constant(str, sh_prelude_class);

 	sl_strlcpy(sh_prelude_version, IDMEF_ANALYZER_VERSION, 32);
	ret = idmef_analyzer_new_version(analyzer, &str);
        prelude_string_set_constant(str, sh_prelude_version);
                
        return 1;
}



static int samhain_alert_prelude(int priority, int sh_class, 
				 char *message, s_modification_t mod,
				 unsigned long msgid)
{
        int ret;
        idmef_time_t *time;
        idmef_alert_t *alert;
        idmef_message_t *idmef;
        idmef_classification_t *classification;
        idmef_assessment_t *assessment;
        idmef_additional_data_t *data;
        idmef_impact_t *impact;
        idmef_target_t *target;
        idmef_confidence_t *confidence;
        prelude_string_t *str;
        
	/*
	 * Currently unused
	 */
	(void) mod;

        if ( sh_class == STAMP)
                return 0;
        
        ret = idmef_message_new(&idmef);
        if ( ret < 0 )
                goto err;
        
        ret = idmef_message_new_alert(idmef, &alert);
        if ( ret < 0 )
                goto err;

        idmef_alert_set_analyzer(alert, idmef_analyzer_ref(prelude_client_get_analyzer(client)), 0);
        
        ret = idmef_time_new_from_gettimeofday(&time);
        idmef_alert_set_detect_time(alert, time);
        
        ret = idmef_time_new_from_gettimeofday(&time);
        idmef_alert_set_create_time(alert, time);
        
        ret = idmef_alert_new_classification(alert, &classification);
        if ( ret < 0 )
                goto err;

        idmef_classification_new_text(classification, &str);
	/*
	 * Classification
	 */
        map_policy_to_class(message, msgid, str);

        //prelude_string_set_ref(str, get_value(message, _("msg"), NULL));
        
        ret = idmef_alert_new_target(alert, &target, -1);
        if ( ret < 0 )
                goto err;

        idmef_target_set_decoy(target, IDMEF_TARGET_DECOY_NO);

	/* where is the node set ???
	 */
	if (NULL != idmef_analyzer_get_node(prelude_client_get_analyzer(client)))
	  {
	    idmef_target_set_node(target, 
				  idmef_node_ref(idmef_analyzer_get_node(prelude_client_get_analyzer(client))));
	  }

        if ( strstr(message, _("path=")) ) {
  
                get_file_infos(target, message, IDMEF_FILE_CATEGORY_ORIGINAL);
                get_file_infos(target, message, IDMEF_FILE_CATEGORY_CURRENT);
        }
        
        ret = idmef_alert_new_assessment(alert, &assessment);
        if ( ret < 0 )
                goto err;
        
        ret = idmef_assessment_new_impact(assessment, &impact);
        
        if ( priority == SH_ERR_SEVERE || priority == SH_ERR_FATAL )
                idmef_impact_set_severity(impact, IDMEF_IMPACT_SEVERITY_HIGH);
        
        else if ( priority == SH_ERR_ALL || priority == SH_ERR_INFO || priority == SH_ERR_NOTICE )
                idmef_impact_set_severity(impact, IDMEF_IMPACT_SEVERITY_LOW);

        else
                idmef_impact_set_severity(impact, IDMEF_IMPACT_SEVERITY_MEDIUM);

        
        ret = idmef_impact_new_description(impact, &str);
        if ( ret < 0 )
                goto err;
        
        if ( sl_strstr(message, _("POLICY")) ) {

                idmef_impact_set_completion(impact, IDMEF_IMPACT_COMPLETION_SUCCEEDED);
                
                if ( sl_strstr(message, _("POLICY KERNEL")) ) {
                        idmef_impact_set_type(impact, IDMEF_IMPACT_TYPE_OTHER);
                        prelude_string_set_dup(str, _("Kernel modification detected by Samhain."));
                } else {
                        idmef_impact_set_type(impact, IDMEF_IMPACT_TYPE_FILE);
                        prelude_string_set_dup(str, _("File system modification detected by Samhain."));
                }
        }

        else {
                idmef_impact_set_completion(impact, IDMEF_IMPACT_COMPLETION_SUCCEEDED);
                
                if ( (sl_strstr(message, _("Login")) ||
                      sl_strstr(message, _("Multiple login")) ||
                      sl_strstr(message, _("Logout"))) && sl_strstr(message, _("Checking")) ) {

                        idmef_impact_set_type(impact, IDMEF_IMPACT_TYPE_USER);
                        prelude_string_set_dup(str, _("Login/logout detected by Samhain."));
                }
                else {
                        idmef_impact_set_type(impact, IDMEF_IMPACT_TYPE_OTHER);
                        prelude_string_set_dup(str, message);
                }
        }

        ret = idmef_assessment_new_confidence(assessment, &confidence);
        if ( ret < 0 )
                goto err;

        idmef_confidence_set_rating(confidence, IDMEF_CONFIDENCE_RATING_HIGH);
        
        ret = idmef_alert_new_additional_data(alert, &data, -1);
        if ( ret < 0 )
                goto err;

        ret = idmef_additional_data_new_meaning(data, &str);
        if ( ret < 0 )
                goto err;

        prelude_string_set_dup(str, _("Message generated by Samhain"));
        idmef_additional_data_set_type(data, IDMEF_ADDITIONAL_DATA_TYPE_STRING);
        idmef_additional_data_set_string_ref(data, message);

        prelude_client_send_idmef(client, idmef);
        idmef_message_destroy(idmef);

        return 0;
        
 err:
        idmef_message_destroy(idmef);
        return -1;
}


int sh_prelude_alert(int priority, int sh_class, char *message,
		     long msgflags, unsigned long msgid)
{
        int ret;
        static int initialized = 0;
	s_modification_t flags = 0;

        /* printf("SH_PRELUDE_ALERT\n"); */
        
	if (msgid == MSG_FI_CHAN)
	  flags = (s_modification_t) msgflags;

        if ( initialized == 0 ) {
                initialized = sh_prelude_init();
                if ( initialized < 0 ) {
                        sh_error_handle((-1), FIL__, __LINE__, -1, MSG_E_SUBGEN,
                                        _("Problem with prelude-ids support: init failed"), 
                                        _("sh_prelude_alert"));
                        return -1;
                }
        }

        if ( initialized < 0 )
                return -1;
        
        ret = samhain_alert_prelude(priority, sh_class, message, flags, msgid);
        if ( ret < 0 ) {
                sh_error_handle((-1), FIL__, __LINE__, -1, MSG_E_SUBGEN,
                                _("Problem with IDMEF for prelude-ids support: alert lost"), 
                                _("sh_prelude_alert"));
        }

        return ret;
}


void sh_prelude_reset(void)
{
  return;
}

/* HAVE_LIBPRELUDE_9 */
#endif

